(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 * TSocketStream Delphi Unit version 0.1                           *
 * Copyright (C) 2005, Matias Moreno                               *
 *                                                                 *
 * Latest version at:                                              *
 * http://www.matiasmoreno.com.ar/libs/delphi/socketstream/        *
 *                                                                 *
 * This library is free software; you can redistribute it and/or   *
 * modify it under the terms of the GNU  Lesser  General  Public   *
 * License  as  published  by  the  Free  Software  Foundation ;   *
 * version 2.1 of the License.                                     *
 *                                                                 *
 * This library is distributed in the hope that it will be useful, *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See  the  *
 * GNU Lesser General Public License for more details.             *
 *                                                                 *
 * You should have  received a  copy of the  GNU  Lesser  General  *
 * Public License along with this library;  if not, write to the   *
 * Free  Software  Foundation,  Inc., 59 Temple Place, Suite 330,  *
 * Boston, MA  02111-1307  USA                                     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)

unit unitSocketStream;

interface

uses Windows, SysUtils, Classes, WinSock;

type
  TSocketStreamError = class(Exception)
  public
    constructor Create;
  end;

  TSocketStream = class;  //forward declaration para clases, segn la ayuda de Delphi

  TOnAcceptEvent = procedure (Sender : TObject; Client : TSocketStream) of object;

  TSocketStream = class(TStream)
  private
    SHandle : TSocket;
    function LookupHost (hostname : string) : Cardinal;
  protected
    procedure SetSize (NewSize : Longint); override;
  public
    {no s qu estoy haciendo ac :-) ver ms tarde...}
    OnAccept : TOnAcceptEvent;
    {Leer datos del socket}
    function Read (var Buffer; Count : Longint) : Longint; override;
    {Escribir datos en el socket}
    function Write (const Buffer; Count : Longint) : Longint; override;
    function Seek (Offset : Longint; Origin : Word) : Longint; override;
    {Conectar a un host remoto}
    constructor Connect (remotehost : String; remoteport : Word);
    {Esperar conexiones}
    constructor Listen (localhost : String; localport : Word);
    {Ver si hay datos disponibles para leer}
    function CanRead : boolean;
    {Ver si hay conexiones entrantes para aceptar}
    function CanAccept : boolean;
    {Aceptar una conexin}
    function Accept : TSocketStream;
    {Informacin de la conexin}
    function LocalHost : string;
    function LocalPort : integer;
    function RemoteHost : string;
    function RemotePort : integer;
    {}
    destructor Destroy; override;
  end;

implementation

constructor TSocketStreamError.Create;
var
  code : Cardinal;
  msg : String;
begin
  code := WSAGetLastError;
  Case code of
  WSAEWOULDBLOCK          : msg := 'WSAEWOULDBLOCK';
  WSAEINPROGRESS          : msg := 'WSAEINPROGRESS';
  WSAEALREADY             : msg := 'WSAEALREADY';
  WSAENOTSOCK             : msg := 'WSAENOTSOCK';
  WSAEDESTADDRREQ         : msg := 'WSAEDESTADDRREQ';
  WSAEMSGSIZE             : msg := 'WSAEMSGSIZE';
  WSAEPROTOTYPE           : msg := 'WSAEPROTOTYPE';
  WSAENOPROTOOPT          : msg := 'WSAENOPROTOOPT';
  WSAEPROTONOSUPPORT      : msg := 'WSAEPROTONOSUPPORT';
  WSAESOCKTNOSUPPORT      : msg := 'WSAESOCKTNOSUPPORT';
  WSAEOPNOTSUPP           : msg := 'WSAEOPNOTSUPP';
  WSAEPFNOSUPPORT         : msg := 'WSAEPFNOSUPPORT';
  WSAEAFNOSUPPORT         : msg := 'WSAEAFNOSUPPORT';
  WSAEADDRINUSE           : msg := 'WSAEADDRINUSE';
  WSAEADDRNOTAVAIL        : msg := 'WSAEADDRNOTAVAIL';
  WSAENETDOWN             : msg := 'WSAENETDOWN';
  WSAENETUNREACH          : msg := 'WSAENETUNREACH';
  WSAENETRESET            : msg := 'WSAENETRESET';
  WSAECONNABORTED         : msg := 'WSAECONNABORTED';
  WSAECONNRESET           : msg := 'WSAECONNRESET';
  WSAENOBUFS              : msg := 'WSAENOBUFS';
  WSAEISCONN              : msg := 'WSAEISCONN';
  WSAENOTCONN             : msg := 'WSAENOTCONN';
  WSAESHUTDOWN            : msg := 'WSAESHUTDOWN';
  WSAETOOMANYREFS         : msg := 'WSAETOOMANYREFS';
  WSAETIMEDOUT            : msg := 'WSAETIMEDOUT';
  WSAECONNREFUSED         : msg := 'WSAECONNREFUSED';
  WSAELOOP                : msg := 'WSAELOOP';
  WSAENAMETOOLONG         : msg := 'WSAENAMETOOLONG';
  WSAEHOSTDOWN            : msg := 'WSAEHOSTDOWN';
  WSAEHOSTUNREACH         : msg := 'WSAEHOSTUNREACH';
  WSAENOTEMPTY            : msg := 'WSAENOTEMPTY';
  WSAEPROCLIM             : msg := 'WSAEPROCLIM';
  WSAEUSERS               : msg := 'WSAEUSERS';
  WSAEDQUOT               : msg := 'WSAEDQUOT';
  WSAESTALE               : msg := 'WSAESTALE';
  WSAEREMOTE              : msg := 'WSAEREMOTE';
  WSAEDISCON              : msg := 'WSAEDISCON';
  WSASYSNOTREADY          : msg := 'WSASYSNOTREADY';
  WSAVERNOTSUPPORTED      : msg := 'WSAVERNOTSUPPORTED';
  WSANOTINITIALISED       : msg := 'WSANOTINITIALISED';
  WSAHOST_NOT_FOUND       : msg := 'WSAHOST_NOT_FOUND';
  WSATRY_AGAIN            : msg := 'WSATRY_AGAIN';
  WSANO_RECOVERY          : msg := 'WSANO_RECOVERY';
  WSANO_DATA              : msg := 'WSANO_DATA / WSANO_ADDRESS';
  else
    msg := 'sin descripcin';
  end;
  msg := 'TSocketStreamError'#13#10+'0x'+IntToHex(Int64(code),8)+': '+msg;
  WSASetLastError (0);
  inherited Create(msg);
end;

function TSocketStream.LookupHost (hostname : string) : Cardinal;
type
  PointerToCardinal = ^Cardinal;
var
  he : PHostEnt;
  x : Cardinal;
begin
  x := Cardinal (inet_addr (PChar(hostname)));
  If x <> INADDR_NONE Then
  begin
    LookupHost := x;
    Exit;
  end;
  he := gethostbyname (PChar(hostname));
  If he = nil Then
  begin
    LookupHost := INADDR_NONE;
    Exit;
  end;
  LookupHost := PointerToCardinal(he^.h_addr_list^)^;
end;

procedure TSocketStream.SetSize (NewSize : Longint);
begin
  //No puedo cambiar el "tamao" de un socket, puesto que no tiene tal cosa
end;

function TSocketStream.Read (var Buffer; Count : Longint) : Longint;
var
  resultado : integer;
begin
  resultado := recv (SHandle, Buffer, Count, 0);
  If resultado = SOCKET_ERROR Then
    raise TSocketStreamError.Create;
  Read := resultado;
end;

function TSocketStream.Write (const Buffer; Count : Longint) : Longint;
var
  resultado : integer;
begin
  //Porque send est declarado con "var Buffer" en vez de "const Buffer" tengo que hacer este arreglito :-(
  resultado := send (SHandle, Pointer(@Buffer)^, Count, 0);
  If resultado = SOCKET_ERROR Then
    raise TSocketStreamError.Create;
  Write := resultado;
end;

function TSocketStream.Seek (Offset : Longint; Origin : Word) : Longint;
begin
  //No puedo desplazarme en un socket
  Seek := 0;
end;

constructor TSocketStream.Connect (remotehost : String; remoteport : Word);
var
  remoteaddr : Cardinal;
  addr : sockaddr_in;
begin
  remoteaddr := LookupHost (remotehost);
  If remoteaddr = INADDR_NONE Then
    raise TSocketStreamError.Create;
  SHandle := socket (AF_INET, SOCK_STREAM, 0);
  If SHandle = INVALID_SOCKET Then
    raise TSocketStreamError.Create;
  FillChar (addr, Sizeof(addr), 0);
  addr.sin_family := AF_INET;
  addr.sin_port := htons(remoteport);
  addr.sin_addr.s_addr := remoteaddr;
  If WinSock.connect (SHandle, addr, sizeof(addr)) = SOCKET_ERROR Then
    raise TSocketStreamError.Create;
end;

constructor TSocketStream.Listen (localhost : String; localport : Word);
var
  localaddr : Cardinal;
  addr : sockaddr_in;
begin
  OnAccept := nil;
  If localhost <> '' Then
  begin
    localaddr := LookupHost (localhost);
    If localaddr = INADDR_NONE Then
      raise TSocketStreamError.Create;
  end else
  begin
    localaddr := INADDR_ANY;
  end;
  SHandle := socket (AF_INET, SOCK_STREAM, 0);
  If SHandle = INVALID_SOCKET Then
    raise TSocketStreamError.Create;
  FillChar (addr, sizeof(addr), 0);
  addr.sin_family := AF_INET;
  addr.sin_port := htons (localport);
  addr.sin_addr.s_addr := localaddr;
  If bind (SHandle, addr, sizeof(addr)) = SOCKET_ERROR Then
    raise TSocketStreamError.Create;
  If WinSock.listen (SHandle, 10) = SOCKET_ERROR Then  //backlog es la cantidad de conexiones pendientes de aceptacin al mismo tiempo... ver la ayuda para ms info.
    raise TSocketStreamError.Create;
end;

destructor TSocketStream.Destroy;
begin
  closesocket (SHandle);
end;

function TSocketStream.CanRead : boolean;
var
  fdset : TFDSet;
  tv : TTimeVal;
begin
  {Ponemos el intervalo de tiempo en cero, para hacer "polling" en vez de esperar}
  tv.tv_sec := 0;
  tv.tv_usec := 0;
  {Hacemos el arreglo de sockets a chequear}
  FD_ZERO (fdset);
  FD_SET (SHandle, fdset);
  {Chequeamos la lista a ver si alguno se puede leer}
  If select (1, @fdset, nil, nil, @tv) = SOCKET_ERROR Then
    raise TSocketStreamError.Create;
  CanRead := FD_ISSET (SHandle, fdset);
end;

function TSocketStream.CanAccept : boolean;
begin
  CanAccept := CanRead; //s, s, exactamente lo mismo! :-)
end;

function TSocketStream.Accept : TSocketStream;
var
  ss : TSocketStream;
  accepthandle : TSocket;
begin
  accepthandle := WinSock.accept (SHandle, nil, nil);
  If accepthandle = INVALID_SOCKET Then
    raise TSocketStreamError.Create;
  ss := TSocketStream.Create;
  ss.SHandle := accepthandle;
  Accept := ss;
end;

function TSocketStream.LocalHost : string;
var
  addr : sockaddr_in;
  addrlen : integer;
begin
  addrlen := sizeof(addr);
  getsockname (SHandle, addr, addrlen);
  LocalHost := inet_ntoa (addr.sin_addr);
end;

function TSocketStream.LocalPort : integer;
var
  addr : sockaddr_in;
  addrlen : integer;
begin
  addrlen := sizeof(addr);
  getsockname (SHandle, addr, addrlen);
  LocalPort := ntohs(addr.sin_port);
end;

function TSocketStream.RemoteHost : string;
var
  addr : sockaddr_in;
  addrlen : integer;
begin
  addrlen := sizeof(addr);
  getpeername (SHandle, addr, addrlen);
  RemoteHost := inet_ntoa (addr.sin_addr);
end;

function TSocketStream.RemotePort : integer;
var
  addr : sockaddr_in;
  addrlen : integer;
begin
  addrlen := sizeof(addr);
  getpeername (SHandle, addr, addrlen);
  RemotePort := ntohs(addr.sin_port);
end;

var
  data : wsadata;
initialization
  If WSAStartup ($0101, data) <> 0 Then
  begin
    MessageBox (0, 'No se pudo inicializar WinSock (versin 1.1)', 'TSocketStream lib error', MB_ICONSTOP);
    Halt;
  end;
finalization
  WSACleanup;
end.
